import os
import sys
import yaml
from subprocess import Popen
import time
import torch
import torchvision
from torchvision.transforms import ToTensor, Lambda
from torch.utils.data import DataLoader
from model import HQS
from wiener import WienerDeconv
import utils
from pypher.pypher import psf2otf
import matplotlib.pyplot as plt
import matplotlib
from matplotlib import rc

plt.rcParams['font.size'] = 14
plt.rcParams["font.family"] = "Arial"

if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"

torch.manual_seed(3452)


max_iters = 5000
nstd_ = 0.1
bstd_ = 2.5
nstep_ = 5
reg_ = 'dncnn'


def convolve_add_noise(params, imgs, cFT, randn, device="cpu"):
    # blurring and adding noise to images
    with torch.no_grad():
        imgs_fft = torch.fft.fft2(imgs)
        blurred_imgs = cFT.unsqueeze(0) * imgs_fft
        input_imgs = torch.real(torch.fft.ifft2(blurred_imgs)) + \
            params["dataset"]["sigma_noise"] * randn
        input_imgs = input_imgs.double()
        return input_imgs


def get_config_params(config_path, dataset, n_std, b_std, n_step, reg):

    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)

    config["dataset"]["name"] = dataset
    config["dataset"]["sigma_blur"] = b_std
    config["dataset"]["sigma_noise"] = n_std
    config["model"]["num_iters"] = n_step
    config["model"]["regularization"] = reg
    config["expname"] = config["expname"] + '_dataset=' + dataset + '_' + 'nstd=' + str(n_std) + \
        '_bstd=' + str(b_std) + '_nstep=' + str(n_step) + '_reg=' + reg
    config['training']['max_iters'] = max_iters

    save_path = "./configs/report_exps"
    filepath = os.path.join(save_path, config["expname"] + '.yaml')
    return filepath, config

def get_model(def_config_path, n_std, b_std, n_step, reg, img_shape):
    _, config = get_config_params(def_config_path, 'mnist', n_std, b_std, n_step, reg)
    model = HQS(config, img_shape=img_shape, device=device)
    ckpt_path = os.path.join("./exps", config["expname"], "%d.pt" % (max_iters))
    ckpt = torch.load(ckpt_path, map_location=device)
    model.load_state_dict(ckpt['model_state_dict'])
    model.to(device)
    return model

@torch.no_grad()
def get_row(gt, n_std, b_std, n_step, reg, img_shape, randn):

    if n_std is None:
        n_std = nstd_
    elif b_std is None:
        b_std = bstd_
    elif n_step is None:
        n_step = nstep_
    elif reg is None:
        reg = reg_

    with torch.no_grad():
        c = utils.fspecial_gaussian_2d(img_shape[1:], b_std)
        cFT = psf2otf(c, (img_shape[1], img_shape[2]))
        cFT = torch.tensor(cFT, device=device)
    params = {"dataset": {"sigma_noise": n_std}}
    input_imgs = convolve_add_noise(params, gt, cFT, randn, device=device)

    # defining models
    vanilla_model = get_model('./configs/default/hqs_vanilla.yaml', n_std, b_std, n_step, reg, img_shape)
    iter_model = get_model('./configs/default/hqs_per_iter.yaml', n_std, b_std, n_step, reg, img_shape)
    image_model = get_model('./configs/default/hqs_per_image.yaml', n_std, b_std, n_step, reg, img_shape)

    vanilla_output = vanilla_model(input_imgs, cFT=cFT).squeeze(0).cpu().numpy()
    iter_output = iter_model(input_imgs, cFT=cFT).squeeze(0).cpu().numpy()
    image_output = image_model(input_imgs, cFT=cFT).squeeze(0).cpu().numpy()

    return input_imgs.squeeze(0).squeeze(0).cpu().numpy(), vanilla_output.squeeze(0), iter_output.squeeze(0), image_output.squeeze(0)


def save_fig(gt, img_shape, var_list, var_name, label, save_path, xlabels=None):

    fig, axes = plt.subplots(nrows=3, ncols=5, figsize=(13, 8))
    plt.subplots_adjust(left=0.1, bottom=0.1, right=0.9, top=0.9, wspace=0.4, hspace=0.4)

    gt = gt.to(device)

    params = {
        'n_std': nstd_,
        'b_std': bstd_,
        'n_step': nstep_,
        'reg': reg_,
        'img_shape': img_shape,
        'randn': torch.randn(*gt.shape, device=device)
    }

    idx = 0

    axes[0, 0].set_title('Input', size='large', weight='bold')
    axes[0, 1].set_title('Vanilla Model', size='large', weight='bold')
    axes[0, 2].set_title('Per-Image Model', size='large', weight='bold')
    axes[0, 3].set_title('Per-Iteration Model', size='large', weight='bold')

    var_list.sort()
    for v in var_list:
        params[var_name] = v
        input_imgs, vanilla_output, iter_output, image_output = get_row(gt, **params)

        axes[idx, 0].imshow(input_imgs, cmap='gray')
        axes[idx, 1].imshow(vanilla_output, cmap='gray')
        axes[idx, 2].imshow(image_output, cmap='gray')
        axes[idx, 3].imshow(iter_output, cmap='gray')

        if xlabels is not None:
            axes[idx, 1].set_xlabel('PSNR %.1f' % (xlabels[idx]["vanilla"]), size='large', weight='bold')
            axes[idx, 2].set_xlabel('PSNR %.1f' % (xlabels[idx]["image"]), size='large', weight='bold')
            axes[idx, 3].set_xlabel('PSNR %.1f' % (xlabels[idx]["iteration"]), size='large', weight='bold')

        axes[idx, 0].set_ylabel('%s=%s' % (label, str(v)), size='large')
        axes[idx, 0].yaxis.set_label_position('left')
        axes[idx, 0].yaxis.tick_right()

        idx += 1

    axes[1, 4].set_title('GT', size='large', weight='bold')
    axes[1, 4].imshow(gt.squeeze(0).squeeze(0).cpu().numpy(), cmap='gray')

    for ax in axes.flatten():
        ax.xaxis.set_ticks([])
        ax.yaxis.set_ticks([])

    # hiding images
    axes[0, 4].remove()
    axes[2, 4].remove()

    plt.tight_layout()
    plt.savefig(save_path)
    plt.clf()


def get_filtering_model(config_path):
    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)
    return WienerDeconv(config)


@torch.no_grad()
def create_main_figure(dataset, params_vanilla, params_iter, params_image, img_shape, save_path, num_samples=7):

    plt.rcParams['font.size'] = 16

    fig, axes = plt.subplots(nrows=num_samples, ncols=7, figsize=(17, 15))
    indices = torch.randint(0, len(dataset), size=(num_samples,))

    vanilla_model = get_model('./configs/default/hqs_vanilla.yaml', **params_vanilla)
    image_model = get_model('./configs/default/hqs_per_image.yaml', **params_vanilla)
    iter_model = get_model('./configs/default/hqs_per_iter.yaml', **params_iter)
    inverse_model = get_filtering_model("./configs/inverse_baseline.yaml")
    wiener_model = get_filtering_model("./configs/wiener_baseline.yaml")

    axes[0, 0].set_title('Input', size='large', weight='bold')
    axes[0, 1].set_title('Vanilla Model', size='large', weight='bold')
    axes[0, 2].set_title('Per-Image Model', size='large', weight='bold')
    axes[0, 3].set_title('Per-Iteration Model', size='large', weight='bold')
    axes[0, 4].set_title("Inverse Deconv", size='large', weight='bold')
    axes[0, 5].set_title("Wiener Deconv", size='large', weight='bold')
    axes[0, 6].set_title("GT", size='large', weight='bold')

    with torch.no_grad():
        c = utils.fspecial_gaussian_2d(img_shape[1:], bstd_)
        cFT = psf2otf(c, (img_shape[1], img_shape[2]))
        cFT = torch.tensor(cFT, device=device)
    params = {"dataset": {"sigma_noise": nstd_}}

    for i, idx in enumerate(indices):
        
        gt, _ = dataset[idx]
        gt = gt.unsqueeze(0)
        gt = gt.to(device)
        randn = torch.randn(*img_shape)

        input_imgs = convolve_add_noise(params, gt, cFT, randn, device=device)

        vanilla_output = vanilla_model(input_imgs, cFT=cFT).squeeze(0).squeeze(0).cpu().numpy()
        image_output = image_model(input_imgs, cFT=cFT).squeeze(0).squeeze(0).cpu().numpy()
        iter_output = iter_model(input_imgs, cFT=cFT).squeeze(0).squeeze(0).cpu().numpy()
        inverse_output = inverse_model(input_imgs, cFT=cFT, sigma=nstd_).squeeze(0).squeeze(0).cpu().numpy()
        wiener_output = wiener_model(input_imgs, cFT=cFT, sigma=nstd_).squeeze(0).squeeze(0).cpu().numpy()
        input_imgs = input_imgs.squeeze(0).squeeze(0).cpu().numpy()

        axes[i, 0].imshow(input_imgs, cmap='gray')
        axes[i, 1].imshow(vanilla_output, cmap='gray')
        axes[i, 2].imshow(image_output, cmap='gray')
        axes[i, 3].imshow(iter_output, cmap='gray')
        axes[i, 4].imshow(inverse_output, cmap='gray')
        axes[i, 5].imshow(wiener_output, cmap='gray')
        axes[i, 6].imshow(gt.squeeze(0).squeeze(0).cpu().numpy(), cmap='gray')

    for ax in axes.flatten():
        ax.xaxis.set_ticks([])
        ax.yaxis.set_ticks([])

    plt.tight_layout()
    plt.savefig(save_path)
    plt.clf()
    

def main():

    os.makedirs("./figs", exist_ok=True)
    dataset = torchvision.datasets.MNIST(root="./mnist", train=False, download=True, transform=ToTensor())
    img_shape = dataset[0][0].shape
    indices = torch.randint(0, len(dataset), size=(4,))

    noise_std = [0.1, 1, 0.01]#, 0.00001]
    psnr_noise = {
        1: {"vanilla": 19.0386, "iteration": 19.2469, "image": 19.1763},
        2: {"vanilla": 12.6915, "iteration": 12.6843, "image": 12.67},
        0: {"vanilla": 24.8461, "iteration": 25.8529, "image": 25.7526}
    }
    save_fig(dataset[indices[0]][0].unsqueeze(0), img_shape, noise_std, 'n_std', r'$\sigma_{noise}$', './figs/noise.png', xlabels=psnr_noise)

    blur_std = [2.5, 1, 5]
    psnr_blur = {
        1: {"vanilla": 19.0386, "iteration": 19.2469, "image": 19.1763},
        0: {"vanilla": 25.4636, "iteration": 25.5074, "image": 25.5458},
        2: {"vanilla": 14.4015, "iteration": 14.4790, "image": 14.3945}
    }
    save_fig(dataset[indices[1]][0].unsqueeze(0), img_shape, blur_std, 'b_std', r'$\sigma_{blur}$', './figs/blur.png', xlabels=psnr_blur)

    num_steps = [5, 1, 10]
    psnr_steps = {
        0: {"vanilla": 14.4812, "iteration": 14.6018, "image": 14.6237},
        1: {"vanilla": 19.0386, "iteration": 19.2469, "image": 19.1763},
        2: {"vanilla": 19.1541, "iteration": 19.3406, "image": 19.3271}
    }
    save_fig(dataset[indices[2]][0].unsqueeze(0), img_shape, num_steps, 'n_step', 'steps', './figs/nsteps.png', xlabels=psnr_steps)

    regs = ['dncnn', 'tv_ani', 'tv_iso']
    psnr_regs = {
        0: {"vanilla": 19.0386, "iteration": 19.2469, "image": 19.1763},
        1: {"vanilla": 14.6589, "iteration": 14.6993, "image": 14.6894},
        2: {"vanilla": 14.6589, "iteration": 14.6993, "image":14.6894}
    }
    save_fig(dataset[indices[3]][0].unsqueeze(0), img_shape, regs, 'reg', 'reg', './figs/reg.png', xlabels=psnr_regs)


    # main comparison figure (best models) with wiener and inverse filtering baselines
    params_vanilla = {
        'n_std': nstd_,
        'b_std': bstd_,
        'n_step': 10,
        'reg': reg_,
        'img_shape': img_shape,
    }
    params_iter = {
        'n_std': nstd_,
        'b_std': bstd_,
        'n_step': 10,
        'reg': reg_,
        'img_shape': img_shape
    }
    params_image = {
        'n_std': nstd_,
        'b_std': bstd_,
        'n_step': 10,
        'reg': reg_,
        'img_shape': img_shape
    }
    create_main_figure(dataset, params_vanilla, params_iter, params_image, img_shape, './figs/main.png', num_samples=5)

if __name__=="__main__":
    main()