
from pypher.pypher import psf2otf
import torch
from torch import nn
import numpy as np

from dncnn_models.dncnn import DnCNN


class HQSPredictor(nn.Module):
    def __init__(self, params, input_dim, device):
        super().__init__()
        self.device = device

        assert not (params['model']['per_image_params'] and params['model']['per_iter_params'])

        # input dim for per_image_params is flatten image dim
        # input dim for per_iter_params is dim(x) + dim(z)
        self.model = nn.Sequential(
            nn.Linear(input_dim, input_dim // 2),
            nn.ReLU(),
            nn.Linear(input_dim // 2, input_dim // 4),
            nn.ReLU(),
            nn.Linear(input_dim // 4, input_dim // 4),
            nn.ReLU(),
            nn.Linear(input_dim // 4,  2)).double()
        
        self.to(self.device)
        

    def forward(self, x):
        return self.model(x.flatten(1))

# HQS code based on Jared's HW6 implementation and provided starter code
class HQS(nn.Module):
    def __init__(self, params, img_shape, device):
        super().__init__()
        self.device = device
        self.eps = torch.tensor([1e-8], device=self.device)
        
        # define blur kernel (trainable or fixed based on options)
        # define trainable rho, lambda, etc
        self.reg = params['model'].get('regularization', 'tv_ani')
        assert self.reg in ['tv_ani', 'tv_iso', 'dncnn']
        self.using_tv = self.reg in ['tv_ani', 'tv_iso']
        self.num_iters = params['model']['num_iters']
        self.per_image_params = params['model'].get('per_image_params', False)
        self.per_iter_params = params['model'].get('per_iter_params', False)

        if self.reg == 'dncnn':
            self.dncnn_model = self.load_dncnn(params['model']['dncnn_path'])

        # initialize trainable parameters
        # use lam and rho properties to read
        if self.per_image_params or self.per_iter_params:
            self.__lam_rho = torch.tensor([[0.1, 0.1]], dtype=torch.float32, device=self.device)
            self.predictor = HQSPredictor(params, img_shape[0] * img_shape[1] * img_shape[2], device)
        else:
            self.__lam_rho = nn.Parameter(data=torch.tensor([0.1, 0.1], dtype=torch.float32))

        # precompute what we can
        if self.using_tv:
            dx = np.array([[-1., 1.]])
            dy = np.array([[-1.], [1.]])
            self.dxFT = torch.tensor(psf2otf(dx, img_shape[1:])).unsqueeze(0).to(self.device)
            self.dyFT = torch.tensor(psf2otf(dy, img_shape[1:])).unsqueeze(0).to(self.device)
            self.dxTFT = torch.conj(self.dxFT).unsqueeze(0).to(self.device)
            self.dyTFT = torch.conj(self.dyFT).unsqueeze(0).to(self.device)
            self.dxyFT = torch.stack((self.dxFT, self.dyFT), axis=0).unsqueeze(0).to(self.device)
            self.dxyTFT = torch.stack((self.dxTFT, self.dyTFT), axis=0).unsqueeze(0).to(self.device)

    @property
    def lam(self):
        return torch.exp(self.__lam_rho[...,0]) + self.eps
    
    @property
    def rho(self):
        return torch.exp(self.__lam_rho[...,1]) + self.eps
    
    def run_predictor(self, x):
        assert self.per_image_params or self.per_iter_params
        self.__lam_rho = self.predictor(x)
        self.__lam_rho = self.__lam_rho.view(-1, 1, 1, 1, 2)

    def load_dncnn(self, model_path):
        model = DnCNN(in_nc=1, out_nc=1, nc=64, nb=17, act_mode='R')        
        model.load_state_dict(torch.load(model_path), strict=True)
        model.eval()
        #for k, v in model.named_parameters():
        #    v.requires_grad = False
        model = model.to(self.device)
        return model


    def hqs_step(self, cFT, cTFT, bFT, x, z):
        # x-update
        if self.using_tv:
            numerator = (cTFT * bFT) + self.rho * (self.dxTFT * torch.fft.fft2(z[:,0,...]) + self.dyTFT * torch.fft.fft2(z[:,1,...]))
            denominator = (cTFT * cFT) + self.rho * (self.dxTFT * self.dxFT + self.dyTFT * self.dyFT)
        else:
            numerator = cTFT * bFT + self.rho * torch.fft.fft2(z)
            denominator = cFT * cTFT + self.rho
        x = torch.fft.ifft2(numerator / denominator)

        # z-update
        if self.using_tv:
            kappa = self.lam / self.rho
            if self.per_image_params or self.per_iter_params:
                kappa = kappa.unsqueeze(1)
            v = torch.fft.ifft2(self.dxyFT * torch.fft.fft2(x).unsqueeze(1))
            if self.reg == 'tv_ani':
                z = torch.maximum(1 - kappa / torch.abs(v), torch.tensor([0], dtype=torch.float32, device=self.device)) * v
            else:
                v = torch.real(v)
                vnorm = torch.sqrt( v[:,0,:,:] ** 2 + v[:,1,:,:] ** 2 )
                if len(kappa.shape) > len(vnorm.shape):
                    kappa = kappa.squeeze(1)
                z = torch.zeros_like(z)
                z[:,0,:,:] = torch.maximum(1 - kappa / vnorm, torch.tensor([0], dtype=torch.float32, device=self.device)) * v[:,0,:,:]
                z[:,1,:,:] = torch.maximum(1 - kappa / vnorm, torch.tensor([0], dtype=torch.float32, device=self.device)) * v[:,1,:,:]
        else:
            # Run denoiser model with shape (B C H W)
            z = torch.zeros_like(z)
            _x = torch.real(x).float()
            # The dncnn only works with one channel, so run each channel separately
            for i in range(_x.shape[1]):
                z[:,i:i+1,...] = self.dncnn_model(_x[:,i:i+1,...])

        return torch.real(x), z


    def forward(self, imgs, cFT, sigma=None):
        '''
        imgs is batched. cFT is not batched
        '''

        cFT = cFT.repeat(imgs.shape[0], 1, 1)
        if len(cFT.shape) != len(imgs.shape):
            cFT = cFT.unsqueeze(1)


        # the following are all batched
        cTFT = torch.conj(cFT)
        bFT = torch.fft.fft2(imgs).to(self.device)

        if self.per_image_params:
            self.run_predictor(imgs)

        x = torch.zeros_like(imgs).to(self.device)
        if self.using_tv:
            z = torch.zeros((imgs.shape[0], 2, *imgs.shape[1:])).to(self.device)
        else:
            z = torch.zeros_like(imgs).to(self.device)

        for _ in range(self.num_iters): 
            if self.per_iter_params:
                self.run_predictor(torch.clip(x, 0, 1))
            x, z = self.hqs_step(cFT, cTFT, bFT, x, z)

        return torch.clip(x, 0, 1).float()
