import torch
import numpy as np
import sys, os
from tqdm import tqdm
import yaml
import torchvision
from torch.utils.data import DataLoader
from model import HQS
from torch import nn
from torch.utils.tensorboard import SummaryWriter
import utils
import yaml
from pypher.pypher import psf2otf
import argparse
from torchvision.transforms import ToTensor, Lambda
from torchmetrics.image import PeakSignalNoiseRatio
from wiener import WienerDeconv

torch.manual_seed(3223)

if torch.cuda.is_available():
    device = "cuda"
else:
    device = "cpu"

def train(params):

    # TODO: implement PSNR for train/val and also testing

    # define training datasets
    if params['dataset']['name'] == "mnist":
        dataset = torchvision.datasets.MNIST(root="./mnist", train=True, download=True, transform=ToTensor())
    elif params['dataset']['name'] == "cifar10":
        dataset = torchvision.datasets.CIFAR10(root="./cifar", train=True, download=True, transform=ToTensor())
    img_shape = dataset[0][0].shape
    
    generator = torch.Generator().manual_seed(42)
    num_train_examples = int(0.9 * len(dataset))
    num_val_examples = int(len(dataset) - num_train_examples)
    train_set, val_set = torch.utils.data.random_split(dataset, [num_train_examples, num_val_examples], \
                                                       generator=generator)

    train_dataloader = DataLoader(train_set, batch_size=params["training"]["batch_size"], shuffle=True)
    val_dataloader = DataLoader(val_set, batch_size=params["training"]["batch_size"], drop_last=True)

    # define hqs model
    model = HQS(params, img_shape=img_shape, device=device)
    model = model.to(device)

    loss_fn = nn.MSELoss()
    metric = PeakSignalNoiseRatio().to(device)

    if params["training"]["optimizer"] == "adam":
        optimizer = torch.optim.Adam(model.parameters(), lr=params['training']['lr'])
    elif params["training"]["optimizer"] == "sgd":
        optimizer = torch.optim.SGD(model.parameters(), lr=params['training']['lr'], \
                                    momentum=params['training']['momentum'])
    elif params["training"]["optimizer"] == "adamw":
        optimizer = torch.optim.AdamW(model.parameters(), lr=params['training']['lr'])
    else:
        raise ValueError("Optimizer %s is not implemented" % (params["training"]["optimizer"]))

    if params['training']['restore_ckpt'] is not None:
        ckpt = torch.load(params['training']['restore_ckpt'])
        model = model.load_state_dict(ckpt['model_state_dict'])
        optim = optim.load_state_dict(ckpt['optim_state_dict'])
        num_iters = ckpt['num_iters']
        expname = ckpt['expname']
        print("restored checkpoint at iteration %d" % (num_iters))
    else:
        num_iters = 0
        expname = params['expname']
        assert expname is not None, "expname cannot be None"
        print("WARNING: starting training at iteration 0, no checkpoint restored")
    num_epochs = params["training"]["num_epochs"]

    if params['training']['scheduler']['use_scheduler']:
        scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=params['training']['scheduler']['first_restart'])
    else:
        scheduler = None

    writer = SummaryWriter(log_dir="./runs/%s" % (expname), purge_step=num_iters)

    save_path = os.path.join("./exps", expname)
    if not os.path.exists(save_path):
        os.makedirs(save_path, exist_ok=True)
        with open(os.path.join(save_path, 'config.yaml'), 'w') as f:
            yaml.dump(params, f, default_flow_style=True)

    with torch.no_grad():
        c = utils.fspecial_gaussian_2d(img_shape[1:], params["dataset"]["sigma_blur"])
        cFT = psf2otf(c, (img_shape[1], img_shape[2]))
        cFT = torch.tensor(cFT, device=device)

    print("starting training....")
    for e in range(num_epochs):
        for _, batch in tqdm(enumerate(train_dataloader)):
            imgs, _ = batch
            imgs = imgs.clone()
            imgs = imgs.to(device)
            input_imgs = utils.convolve_add_noise(params, imgs, cFT, device=device)

            preds = model(input_imgs, cFT=cFT)

            optimizer.zero_grad()
            loss = loss_fn(preds, imgs)
            loss.backward()
            optimizer.step()

            if scheduler is not None:
                scheduler.step()

            with torch.no_grad():
                metric.update(torch.clip(preds, 0, 1), torch.clip(imgs, 0, 1))
        
            num_iters += 1
            if num_iters % params['logging']['log_freq'] == 0:
                writer.add_scalar("train/loss", loss.cpu().item(), num_iters)
                train_psnr = metric.compute().cpu().item()
                writer.add_scalar("train/psnr", train_psnr, num_iters)
                print("iteration %d, loss %.5f, psnr %.3f" % (num_iters, loss.cpu().item(), train_psnr))

            if num_iters % params["logging"]["save_img"] == 0:
                with torch.no_grad():
                    idx = torch.randint(params["training"]["batch_size"], size=(2,), device=device)
                    viz_gt = imgs[idx]
                    viz_input = input_imgs[idx]
                    viz_output = preds[idx]
                    grid = torch.cat([viz_input, viz_output, viz_gt], dim=-1)
                    grid = torchvision.utils.make_grid(grid, nrow=1)
                    writer.add_image("train/imgs", grid, num_iters)

            if num_iters % params['logging']['save_freq'] == 0:
                val_loss, val_psnr = utils.validate_model(params, cFT, model, val_dataloader, loss_fn, device=device)
                print("validation loss at iteration %d: %.3f" % (num_iters, val_loss))
                print("validation psnr at iteration %d: %.3f" % (num_iters, val_psnr))
                writer.add_scalar("val/loss", val_loss, num_iters)
                writer.add_scalar("val/psnr", val_psnr, num_iters)
                utils.save_checkpoint(save_path, expname, model, optimizer, num_iters)

            if num_iters >= params['training']['max_iters']:
                break
        
        if num_iters >= params['training']['max_iters']:
            break

        metric.reset()

    utils.save_checkpoint(save_path, expname, model, optimizer, num_iters)


if __name__=="__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--config-file", type=str, required=True)
    parser.add_argument("--train", action='store_true')
    parser.add_argument("--test", action='store_true')
    args = parser.parse_args()

    with open(args.config_file, 'r') as f:
        config = yaml.safe_load(f)

    if args.train:
        train(config)
    else:
        print("running inference on test set")
        if config['dataset']['name'] == "mnist":
            dataset = torchvision.datasets.MNIST(root="./mnist", train=False, download=True, transform=ToTensor())
        elif config['dataset']['name'] == "cifar10":
            dataset = torchvision.datasets.CIFAR10(root="./cifar", train=False, download=True, transform=ToTensor())
        img_shape = dataset[0][0].shape
        dataloader = DataLoader(dataset, batch_size=config["test"]["batch_size"], drop_last=True)

        if config["test"]["model"] in ("wiener", "inverse"):
            print("loading %s deconvolution" % (config["test"]["model"]))
            model = WienerDeconv(config)
        else:
            print("loading HQS model")
            model = HQS(config, img_shape=img_shape, device=device)
            if config['test']['ckpt'] is not None:
                ckpt = torch.load(config['test']['ckpt'])
                model.load_state_dict(ckpt['model_state_dict'])
            model = model.to(device)
            model.eval()

        with torch.no_grad():
            c = utils.fspecial_gaussian_2d(img_shape[1:], config["dataset"]["sigma_blur"])
            cFT = psf2otf(c, (img_shape[1], img_shape[2]))
            cFT = torch.tensor(cFT).to(device)

        loss_fn = nn.MSELoss()
        test_loss, test_psnr = utils.validate_model(config, cFT, model, dataloader, loss_fn, \
            device=device, sigma=config["dataset"]["sigma_noise"], save_img=config['test']['save_img'])
        print("Test loss: %.4f, test psnr: %.4f" % (test_loss, test_psnr))

        sys.exit(0)