import os
import sys
import yaml
from subprocess import Popen
import time

datasets = ['mnist']#, 'cifar']
noise_std = [0.1, 1, 0.01]#, 0.00001]
blur_std = [2.5, 1, 5]
num_steps = [5, 1, 10]
regs = ['dncnn', 'tv_ani', 'tv_iso']
max_iters = 5000


def set_config_params(config_path, dataset, n_std, b_std, n_step, reg):

    with open(config_path, 'r') as f:
        config = yaml.safe_load(f)

    config["dataset"]["name"] = dataset
    config["dataset"]["sigma_blur"] = b_std
    config["dataset"]["sigma_noise"] = n_std
    config["model"]["num_iters"] = n_step
    config["model"]["regularization"] = reg
    config["expname"] = config["expname"] + '_dataset=' + dataset + '_' + 'nstd=' + str(n_std) + \
        '_bstd=' + str(b_std) + '_nstep=' + str(n_step) + '_reg=' + reg
    config['training']['max_iters'] = max_iters

    save_path = "./configs/report_exps"
    filepath = os.path.join(save_path, config["expname"] + '.yaml')
    os.makedirs(save_path, exist_ok=True)
    with open(filepath, 'w') as f:
        yaml.dump(config, f, indent=4)
    return filepath


def run_command(config_path):
    return Popen(["python", "run.py", "--train", "--config-file", config_path]) 


def run_exp(def_config_path, dataset, n_std, b_std, n_step, reg):
    config_path = set_config_params(def_config_path, dataset, n_std, b_std, n_step, reg)
    return run_command(config_path)

def run_exp_set(failed_runs, params, method=None):
    
    # processes running async
    if method is None:
        print('Running all methods with:', params)
        ret1 = run_exp('./configs/default/hqs_vanilla.yaml', **params)
        ret2 = run_exp('./configs/default/hqs_per_iter.yaml', **params)
        ret3 = run_exp('./configs/default/hqs_per_image.yaml', **params)
        proc_list = [ret1, ret2, ret3]
    else:
        print(f'Running {method} with:', params)
        if method == 'vanilla':
            proc_list = [run_exp('./configs/default/hqs_vanilla.yaml', **params)]
        elif method == 'per_iter':
            proc_list = [run_exp('./configs/default/hqs_per_iter.yaml', **params)]
        elif method == 'per_image':
            proc_list = [run_exp('./configs/default/hqs_per_image.yaml', **params)]

    # based on https://stackoverflow.com/questions/636561/how-can-i-run-an-external-command-asynchronously-from-python
    while proc_list:
        for proc in proc_list:
            retcode = proc.poll()
            if retcode is not None:
                # process completed
                proc_list.remove(proc)
                break
            else:
                time.sleep(.1)
                continue

        if retcode is not None and retcode != 0:
            print(proc.args)
            failed_runs.append(proc.args)

def main():

    failed_runs = []

    # First element in each array is the default setting
    # Rather than running every combo, just vary each parameter independently

    # First run all defaults
    defaults = {
        'dataset': datasets[0],
        'n_std': noise_std[0],
        'b_std': blur_std[0],
        'n_step': num_steps[0],
        'reg': regs[0]
    }
    run_exp_set(failed_runs, defaults, method='vanilla')
    run_exp_set(failed_runs, defaults, method='per_iter')
    run_exp_set(failed_runs, defaults, method='per_image')

    # Now vary each parameter independently

    for x in datasets[1:]:
        run_exp_set(failed_runs, defaults | { 'dataset': x })

    for x in noise_std[1:]:
        run_exp_set(failed_runs, defaults | { 'n_std': x })
        
    for x in blur_std[1:]:
        run_exp_set(failed_runs, defaults | { 'b_std': x })

    for x in num_steps[1:]:
        # don't parallelize because they can be super ultra slow
        run_exp_set(failed_runs, defaults | { 'n_step': x }, method='vanilla')
        run_exp_set(failed_runs, defaults | { 'n_step': x }, method='per_iter')
        run_exp_set(failed_runs, defaults | { 'n_step': x }, method='per_image')

    for x in regs[1:]:
        run_exp_set(failed_runs, defaults | { 'reg': x })

                        
    print("The following runs failed:\n")
    for run in failed_runs:
        print(run + "\n")


if __name__=="__main__":
    main()