import torch
import numpy as np
import os
from skimage.filters import gaussian
from tqdm import tqdm
from torchmetrics.image import PeakSignalNoiseRatio
import torchvision

def save_checkpoint(save_path, expname, model, optim, num_iters):
    data = {
        "expname": expname,
        "model_state_dict": model.state_dict(),
        "optim_state_dict": optim.state_dict(),
        "num_iters": num_iters
    }
    save_file = os.path.join(save_path, "%d.pt" % (num_iters))
    torch.save(data, save_file)


# taken from hw6_task2.py
# helper function for computing a 2D Gaussian convolution kernel
def fspecial_gaussian_2d(size, sigma):
    kernel = np.zeros(tuple(size))
    kernel[size[0]//2, size[1]//2] = 1
    kernel = gaussian(kernel, sigma)
    return kernel/np.sum(kernel)


def convolve_add_noise(params, imgs, cFT, device="cpu"):
    # blurring and adding noise to images
    with torch.no_grad():
        imgs_fft = torch.fft.fft2(imgs)
        blurred_imgs = cFT.unsqueeze(0) * imgs_fft
        input_imgs = torch.real(torch.fft.ifft2(blurred_imgs)) + \
            params["dataset"]["sigma_noise"] * torch.randn(*blurred_imgs.shape, device=device)
        input_imgs = input_imgs.double()
        return input_imgs
    
@torch.no_grad()
def validate_model(params, cFT, model, dataloader, loss_fn, device="cpu", sigma=None, save_img=None):
    print("running validation")

    model.eval()
    val_loss = 0.0
    num_batches = 0.0

    metric = PeakSignalNoiseRatio().to(device)

    for i, batch in tqdm(enumerate(dataloader)):
        imgs = batch[0]
        imgs = imgs.to(device)
        input_imgs = convolve_add_noise(params, imgs, cFT, device=device)
        preds = model(input_imgs, cFT=cFT, sigma=sigma)
        val_loss += loss_fn(preds, imgs).cpu().item()
        num_batches += 1.0

        metric.update(torch.clip(preds, 0, 1), torch.clip(imgs, 0, 1))

        if save_img is not None and i % save_img == 0:
            grid = torch.cat([input_imgs, preds, imgs], dim=-1)
            grid = torchvision.utils.make_grid(grid, nrow=1)
            folder = os.path.join('./test', params['expname'])
            os.makedirs(folder, exist_ok=True)
            torchvision.utils.save_image(grid.cpu(), os.path.join(folder, '%d.png' % (i)))

    val_loss = val_loss / num_batches
    model.train()
    psnr = metric.compute().cpu().item()

    return val_loss, psnr